from django.contrib import admin
from django.conf import settings
from django.db.models import QuerySet
from django.http import HttpRequest

from rest_framework.authtoken.models import Token
from rest_framework.authtoken.admin import TokenAdmin

from django_tasks import models
from django_tasks.admin_tools import AdminTaskAction, StatusDisplayModelAdmin
from django_tasks.serializers import DocTaskSerializer


class AdminSite(admin.AdminSite):
    site_title = 'Stored Tasks'
    site_header = 'Stored Tasks'
    index_title = 'Index'


site = AdminSite()
site.register(Token, TokenAdmin)


@AdminTaskAction('django_tasks.tasks.doctask_access_test', description='Test async database access')
def doctask_access_test(modeladmin: admin.ModelAdmin, request: HttpRequest, queryset: QuerySet):
    pass


@AdminTaskAction('django_tasks.tasks.doctask_deletion_test', description='Test async database DELETE')
def doctask_deletion_test(modeladmin: admin.ModelAdmin, request: HttpRequest, queryset: QuerySet):
    pass


@admin.register(models.DocTask, site=site)
class DocTaskModelAdmin(StatusDisplayModelAdmin):
    list_display = ('name', 'inputs', 'duration', *DocTaskSerializer.Meta.read_only_fields)

    if settings.DEBUG:
        actions = [doctask_access_test, doctask_deletion_test]

    def has_change_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request):
        return False



import asyncio
import collections
import logging

from typing import Any

from channels.db import database_sync_to_async
from django_tasks.serializers import DocTaskSerializer
from django_tasks.task_runner import TaskRunner


class DocTaskScheduler:
    doctask_index: dict[int, dict[str, Any]] = collections.defaultdict(dict)
    model = DocTaskSerializer.Meta.model

    @classmethod
    def retrieve_doctask(cls, memory_id):
        if memory_id in cls.doctask_index:
            doctask_info = cls.doctask_index[memory_id]

            try:
                return cls.model.objects.get(id=doctask_info['id'])
            except cls.model.DoesNotExist:
                count = cls.model.objects.count()
                logging.getLogger('django').error(
                    'Memorized doctask ID %s not found in DB, among %s entries.', doctask_info, count)

    @classmethod
    async def store_doctask_result(cls, task_info):
        memory_id = task_info['memory-id']
        doctask = await database_sync_to_async(cls.retrieve_doctask)(memory_id)

        if doctask:
            await doctask.on_completion(TaskRunner.get_task_info(cls.doctask_index[memory_id]['future']))
            del cls.doctask_index[memory_id]
            logging.getLogger('django').info('Stored %s.', repr(doctask))

    @classmethod
    async def schedule_doctask(cls, data={}, callable=None) -> asyncio.Future:
        """Schedules a single task, and stores results in DB."""
        callable = callable or DocTaskSerializer.get_coro_info(data).callable
        runner = TaskRunner.get()
        task = await runner.schedule(callable(**data.get('inputs', {})), cls.store_doctask_result)
        cls.doctask_index[id(task)].update({'future': task, 'id': data.get('id')})
        logging.getLogger('django').info('Scheduled doc-task %s callable=%s.', data, callable)
        return task

    @staticmethod
    async def schedule_task(data={}, callable=None) -> asyncio.Future:
        """Schedules a single task without storage."""
        callable = callable or DocTaskSerializer.get_coro_info(data).callable
        runner = TaskRunner.get()
        task = await runner.schedule(callable(**data.get('inputs', {})))
        logging.getLogger('django').info('Scheduled task %s callable=%s.', data, callable)
        return task



# Generated by Django 4.2.1 on 2023-05-25 17:41

import datetime
from django.db import migrations, models
import django_tasks.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='DocTask',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=80)),
                ('scheduled_at', models.DateTimeField(default=datetime.datetime.now)),
                ('completed_at', models.DateTimeField(null=True)),
                ('inputs', models.JSONField(default=dict, encoder=django_tasks.models.DefensiveJsonEncoder)),
                ('document', models.JSONField(default=list, encoder=django_tasks.models.DefensiveJsonEncoder)),
            ],
        ),
    ]



import functools
import json
import os

from typing import Any, Callable, Optional

import websocket

from django.conf import settings
from django.contrib import admin, messages
from django.db.models import QuerySet
from django.http import HttpRequest


PROXY_ROUTE = settings.CHANNEL_TASKS.proxy_route


class AdminTaskAction:
    header: dict[str, str] = {'Content-Type': 'application/json'}
    timeout = 20

    def __init__(self, task_name: str, **kwargs):
        self.task_name = task_name
        self.kwargs = kwargs
    def __call__(self, post_schedule_callable: Callable[[Any, HttpRequest, QuerySet], Any]):
        @admin.action(**self.kwargs)
        @functools.wraps(post_schedule_callable)
        def action_callable(modeladmin: admin.ModelAdmin, request: HttpRequest, queryset):
            ws_response = self.websocket_task_schedule(
                request, self.task_name, instance_ids=list(queryset.values_list('pk', flat=True))
            )
            modeladmin.message_user(request, ws_response, messages.INFO)
            return post_schedule_callable(modeladmin, request, queryset)

        return action_callable

    def websocket_task_schedule(self, http_request: HttpRequest, task_name: str, **inputs):
        secure = 's' if http_request.is_secure() else ''
        origin = f'http{secure}://{http_request.get_host()}'
        address = os.path.join(http_request.get_host(), PROXY_ROUTE, 'tasks')
        ws = websocket.WebSocket()
        ws.connect(
            f'ws{secure}://{address}/', header=self.header, timeout=self.timeout,
            origin=origin, cookie=http_request.headers.get('Cookie'),
            http_proxy_host=settings.CHANNEL_TASKS.proxy_host,
            http_proxy_port=settings.CHANNEL_TASKS.proxy_port,
            proxy_type=settings.CHANNEL_TASKS.proxy_type,
        )
        ws.send(json.dumps([dict(name=task_name, inputs=inputs)], indent=4))
        response = ws.recv()

        return response


class ExtraContextModelAdmin(admin.ModelAdmin):
    def changelist_view(self, request: HttpRequest, extra_context: Optional[dict] = None):
        extra_context = extra_context or {}
        self.add_changelist_extra_context(request, extra_context)

        return super().changelist_view(request, extra_context=extra_context)

    def add_changelist_extra_context(self, request: HttpRequest, extra_context: dict):
        raise NotImplementedError


class StatusDisplayModelAdmin(ExtraContextModelAdmin):
    change_list_template = 'task_status_display.html'

    def add_changelist_extra_context(self, request: HttpRequest, extra_context: dict):
        extra_context['websocket_uri'] = os.path.join('/', PROXY_ROUTE, 'tasks/')



import asyncio
import time

from rest_framework import status

from . import base


def teardown_module():
    """
    Called by Pytest at teardown of the test module, employed here to
    log final scenario results
    """
    base.BddTester.gherkin.log()


class RestApiWithTokenAuth(base.BddTester):
    """
    Staff users may obtain a token through Django admin site, and use it to schedule
    concurrent tasks through REST API.
    This covers:
    * The task runner
    * Admin site usage to create API tokens
    * User creation with management command
    * The tasks REST API
    """

    @base.BddTester.gherkin()
    def a_user_creates_an_api_token(self):
        """
        Given a tasks admin `user` is created with command
        And the user creates an API `token`
        """

    @base.BddTester.gherkin()
    def test_many_tasks_execution_post_with_result_storage(self):
        """
        Given a user creates an API token
        When a failed and some OK `tasks` are posted
        Then $(0) cancelled $(1) error $(4) success messages are broadcasted
        And the different task results are correctly stored in DB
        """

    @base.BddTester.gherkin()
    def single_task_execution_post_with_result_storage(self):
        """
        Given a user creates an API token
        When a failed `task` is posted with duration $(0.1)
        Then $(0) cancelled $(1) error $(0) success messages are broadcasted
        Then the task result is correctly stored in DB
        """

    async def a_failed_and_some_ok_tasks_are_posted(self):
        task_data = [dict(name='sleep_test', inputs={'duration': dn}) for dn in self.task_durations]
        task_data.append(dict(name='sleep_test', inputs={'duration': 0.15, 'raise_error': True}))
        response = await self.assert_rest_api_call(
            'POST', 'tasks/schedule', status.HTTP_201_CREATED, json_data=task_data)
        time.sleep(max(self.task_durations))

        return response.json(),

    async def the_different_task_results_are_correctly_stored_in_db(self):
        response = await self.assert_rest_api_call('GET', 'tasks', status.HTTP_200_OK)
        tasks = response.json()
        assert len(tasks) == 5
        assert tasks == []

    async def a_failed_task_is_posted_with_duration(self):
        duration = float(self.param)
        data = dict(name='sleep_test', inputs={'duration': duration, 'raise_error': True})
        response = await self.assert_rest_api_call('POST', 'tasks', status.HTTP_201_CREATED, json_data=data)
        response_json = response.json()
        del response_json['scheduled_at']
        assert response_json == {**data, 'completed_at': None, 'document': None}

        return response_json,

    async def the_user_creates_an_api_token(self):
        response = await self.admin_client.post('/authtoken/token/add/', {'user': self.get_output('user').pk})
        redirected_response = await self.assert_admin_redirection(response)
        soup = self.get_soup(redirected_response)
        messages = self.get_all_admin_messages(soup)
        assert len(messages['success']) == 1, soup

        return messages['success'][0].split()[2].strip('“”'),

    async def the_task_result_is_correctly_stored_in_db(self):
        response = await self.assert_rest_api_call('GET', 'tasks', status.HTTP_200_OK)
        tasks = response.json()
        assert len(tasks) == 1


class TestTaskRunner(base.BddTester):
    """
    Several tasks may be scheduled to run concurrently, and their states are broadcasted.
    Task information may also be stored in database.
    This covers:
    * The task runner
    * The websocket broadcasting
    """

    @base.BddTester.gherkin()
    def test_concurrent_error_and_cancellation(self):
        """
        When a `failed`, a `cancelled` and some `OK` tasks are scheduled
        Then completion times do not accumulate
        And $(1) cancelled $(1) error $(4) success messages are broadcasted
        And the different task statuses are correctly stored
        """

    async def a_failed_a_cancelled_and_some_ok_tasks_are_scheduled(self):
        failed_task, cancelled_task, *ok_tasks = await asyncio.gather(
            self.runner.schedule(self.fake_task_coro_raise(0.1)),
            self.runner.schedule(self.fake_task_coro_ok(10)),
            *[self.runner.schedule(self.fake_task_coro_ok(d)) for d in self.task_durations])

        return failed_task, cancelled_task, ok_tasks

    async def completion_times_do_not_accumulate(self):
        initial_time = time.time()
        task_results = await asyncio.gather(*self.get_output('ok'))
        self.get_output('cancelled').cancel()
        elapsed_time = time.time() - initial_time

        assert task_results == self.task_durations
        assert elapsed_time < 1

    async def the_different_task_statuses_are_correctly_stored(self):
        failed_task_info = self.runner.get_task_info(self.get_output('failed'))
        assert failed_task_info['status'] == 'Error'
        assert failed_task_info['traceback'][-1].strip() == 'Exception: Fake error'

        await asyncio.sleep(0.01)
        cancelled_task_info = self.runner.get_task_info(self.get_output('cancelled'))
        assert cancelled_task_info['status'] == 'Cancelled'


class TestWebsocketScheduling(base.BddTester):
    """
    This covers:
    * The task runner
    * The tasks websocket API
    """

    @base.BddTester.gherkin()
    def test_several_tasks_are_scheduled_with_ws_message(self):
        """
        When a failed and some OK tasks are scheduled through WS
        Then $(0) cancelled $(1) error $(4) success messages are broadcasted
        """

    async def a_failed_and_some_ok_tasks_are_scheduled_through_ws(self):
        task_data = [dict(name='sleep_test', inputs={'duration': dn}) for dn in self.task_durations]
        task_data.append(dict(name='sleep_test', inputs={'duration': 0.15, 'raise_error': True}))
        await self.communicator.send_json_to(task_data)


class TestAsyncAdminSiteActions(RestApiWithTokenAuth):
    """
    This covers:
    * The admin tools module
    """

    @base.BddTester.gherkin()
    def test_database_access_async_actions_run_ok(self):
        """
        Given single task execution post with result storage
        When the user runs the $(database_access_test) action
        And the user runs the $(store_database_access_test) action
        And the user runs the $(delete_test) action
        """

    async def the_user_runs_the_action(self):
        response = await self.admin_client.post('/django_tasks/doctask/', {
            'action': self.param,
            '_selected_action': [doctask.pk for doctask in self.models.DocTask.objects.all()]})

        assert response.status_code == status.HTTP_200_OK



import os
import sys


def manage_channel_tasks():
    os.environ['DJANGO_SETTINGS_MODULE'] = 'django_tasks.settings.base'

    from django.core.management import execute_from_command_line

    execute_from_command_line(sys.argv)



import configparser
import setuptools


ini = configparser.ConfigParser()
ini.read('version.ini')


with open('README.md') as readme:
    long_description = readme.read()


tests_require = ['pytest', 'pytest-cov', 'pytest-django', 'pytest-asyncio', 'beautifulsoup4',
                 'bdd-coder==2.2.3.dev2']

setuptools.setup(
    name=ini['version']['name'],
    version=ini['version']['value'],
    author='Daniel Farré Manzorro',
    author_email='d.farre.m@gmail.com',
    description='Running background tasks through REST API and websocket, with channels-redis',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://bitbucket.org/coleopter/django-tasks',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers'],
    packages=setuptools.find_packages(),
    package_data={'django_tasks': ['templates/*', 'settings/channel-task-defaults.ini']},
    data_files=[('./channel-tasks-docker', ['docker-compose.yml',
                                            'Dockerfile',
                                            'setup-channel-tasks-docker.sh',
                                            'channel-tasks-unit.template.json']),
                ('./services', ['services/channel-tasks-docker.template.service',
                                'services/channel-tasks.template.service']),
                ('./bin', ['services/restart-channel-tasks-docker.sh',
                           'services/setup-channel-tasks.sh'])],
    entry_points={'console_scripts': ['channel-tasks-admin=django_tasks.entrypoint:manage_channel_tasks']},
    install_requires=[
        'Django', 'django-filter', 'django-extensions', 'django-request-logging', 'djangorestframework',
        'django-bootstrap-v5', 'channels', 'channels-redis', 'tzdata', 'psycopg2',
        'websocket-client'],
    extras_require={'dev': ['ipdb', 'ipython'],
                    'mypy': ['mypy', 'django-stubs', 'djangorestframework-stubs[compatible-mypy]',
                             'types-beautifulsoup4', 'types-setuptools'],
                    'test': tests_require},
    tests_require=tests_require,
)



import configparser
import os


class SettingsIni:
    ini_key = 'CHANNEL_TASKS_INI_PATH'

    def __init__(self):
        assert self.ini_key in os.environ, f'Settings are to be specified with the {ini_key} envvar.'
        self.ini = configparser.ConfigParser()
        self.ini.read(os.environ[self.ini_key])

    def get_array(self, section, key, default):
        return ([line.strip() for line in self.ini[section][key].splitlines() if line.strip()]
                if self.ini.has_option(section, key) else default)

    def get_boolean(self, section, key, default):
        return self.ini[section].getboolean(key, default) if self.ini.has_section(section) else default

    def get_int(self, section, key, default):
        return self.ini[section].getint(key, default) if self.ini.has_section(section) else default

    def get_text(self, section, key, default):
        return self.ini[section][key].strip() if self.ini.has_option(section, key) else default

    @property
    def allowed_hosts(self):
        return self.get_array('security', 'allowed-hosts', ['localhost'])

    @property
    def install_apps(self):
        return self.get_array('apps', 'install-apps', [])

    @property
    def debug(self):
        return self.get_boolean('security', 'debug', False)

    @property
    def proxy_route(self):
        return self.get_text('security', 'proxy-route', '')

    @property
    def proxy_host(self):
        return self.get_text('security', 'proxy-host', '127.0.0.1')

    @property
    def proxy_port(self):
        return self.get_int('security', 'proxy-port', 8001)

    @property
    def proxy_type(self):
        return self.get_text('security', 'proxy-type', 'http')

    @property
    def log_level(self):
        return self.get_text('logging', 'log-level', 'INFO')

    @property
    def expose_doctask_api(self):
        return self.get_boolean('asgi', 'expose-doctask-api', False)

    @property
    def databases(self):
        if not self.ini.has_section('database'):
            return {
                'default': {
                    'ENGINE': 'django.db.backends.sqlite3',
                    'NAME': 'channel-tasks.sqlite3',
                }
            }

        default = {k.upper(): v for k, v in dict(self.ini['database']).items()}
        default.setdefault('PASSWORD', os.getenv('CHANNEL_TASKS_DB_PASSWORD', ''))

        return {'default': default}



import collections
import inspect
import importlib

from typing import Callable, Any


class TaskCoroInfo:
    _safe_coroutine_modules: set[str] = set()

    @classmethod
    def mark_as_safe(cls, callable: Callable):
        assert inspect.iscoroutinefunction(callable)
        cls._safe_coroutine_modules.add(inspect.getmodule(callable).__spec__.name)
        return callable

    def __init__(self, dotted_path: str, **inputs: dict[str, Any]):
        self.dotted_path = dotted_path.strip()
        self.inputs = inputs
        self.module_path, self.name, self.callable = '', '', None
        self.errors: dict[str, list[str]] = collections.defaultdict(list)
        self.check()

    def check(self):
        if '.' in self.dotted_path:
            self.module_path, self.name = self.dotted_path.rsplit('.', 1)

        if self.module_path in self._safe_coroutine_modules:
            self.check_coroutine()

        if self.callable:
            self.check_inputs()
        else:
            self.errors['name'].append(f"Coroutine '{self.dotted_path}' not found.")

    def check_coroutine(self):
        try:
            module = importlib.import_module(self.module_path)
        except ImportError:
            module = None

        self.callable = getattr(module, self.name, None)

        if not inspect.iscoroutinefunction(self.callable):
            self.callable = None

    def check_inputs(self):
        params = inspect.signature(self.callable).parameters
        required_keys = set(k for k, v in params.items() if v.default == inspect._empty)
        optional_keys = set(k for k, v in params.items() if v.default != inspect._empty)

        input_keys = set(self.inputs)
        missing_keys = required_keys - input_keys
        unknown_keys = input_keys - required_keys - optional_keys

        if missing_keys:
            self.errors['inputs'].append(f'Missing required parameters {missing_keys}.')

        if unknown_keys:
            self.errors['inputs'].append(f'Unknown parameters {unknown_keys}.')



import django

django.setup()

from django import urls
from django.conf import settings
from django.core.asgi import get_asgi_application

from channels.auth import AuthMiddlewareStack
from channels.routing import ProtocolTypeRouter, URLRouter
from channels.security.websocket import AllowedHostsOriginValidator

from rest_framework import routers

from django_tasks.consumers import TaskEventsConsumer, TasksRestConsumer
from django_tasks.viewsets import TaskViewSet


class OptionalSlashRouter(routers.SimpleRouter):
    def __init__(self):
        super().__init__()
        self.trailing_slash = '/?'


http_paths = []


if settings.CHANNEL_TASKS.expose_doctask_api is True:
    drf_router = OptionalSlashRouter()
    drf_router.register('tasks', TaskViewSet)
    http_paths.append(urls.re_path(r'^api/', URLRouter(TasksRestConsumer.get_urls(drf_router))))


http_paths.append(urls.re_path(r'^', get_asgi_application()))
url_routers = {
    'http': URLRouter(http_paths),
    'websocket': AllowedHostsOriginValidator(AuthMiddlewareStack(
        URLRouter([urls.path('tasks/', TaskEventsConsumer.as_asgi())])
    )),
}
application = ProtocolTypeRouter(url_routers)



import asyncio

from typing import Any

from rest_framework import exceptions, serializers

from django_tasks import models
from django_tasks import task_inspector
from django_tasks import task_runner


class DocTaskSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.DocTask
        read_only_fields = ('id', 'scheduled_at', 'completed_at', 'document')
        fields = ('name', 'inputs', *read_only_fields)

    @classmethod
    async def schedule_task_group(cls, json_content, *args, **kwargs):
        kwargs.update(dict(many=True, data=json_content))
        many_serializer = cls(*args, **kwargs)
        many_serializer.is_valid(raise_exception=True)
        runner = task_runner.TaskRunner.get()
        tasks = await asyncio.gather(*[
            runner.schedule(many_serializer.child.get_coro_info(task).callable(**task['inputs']))
            for task in many_serializer.data
        ])
        return many_serializer, tasks

    @classmethod
    def create_doctask_group(cls, json_content, *args, **kwargs):
        kwargs.update(dict(many=True, data=json_content))
        many_serializer = cls(*args, **kwargs)
        many_serializer.is_valid(raise_exception=True)
        doctasks = many_serializer.save()

        return many_serializer, doctasks

    @staticmethod
    def get_coro_info(attrs: dict[str, Any]) -> task_inspector.TaskCoroInfo:
        coro_info = task_inspector.TaskCoroInfo(attrs['name'], **attrs['inputs'])

        if coro_info.errors:
            raise exceptions.ValidationError(coro_info.errors)

        return coro_info

    def validate(self, attrs: dict[str, Any]) -> dict[str, Any]:
        self.context['coro_info'] = self.get_coro_info(attrs)

        return attrs

    def create(self, validated_data: dict[str, Any]) -> models.DocTask:
        """Creates a `DocTask` instance to run the specified function with given arguments.
        The resulting task (actually an `asyncio.Future`) should return a JSON-serializable object
        as result -task document- to be stored; 'inputs' should be valid keyword arguments to `callable`.
        """
        return self.Meta.model.objects.create(name=validated_data['name'], inputs=validated_data['inputs'])
